{-

Final Project - Bingo

Deepakraj Dharmapuri Selvakumar - dd3068

-}

{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE EmptyDataDecls #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE UndecidableInstances #-}

import qualified Data.Map.Strict as Map
import qualified Data.List as List
import Data.List.Split as Split
import qualified Data.Matrix as Mat
import System.Random (newStdGen, RandomGen, mkStdGen)
import System.Random.Shuffle (shuffle')
import Control.Monad(forM)
import System.Exit(die)
import Control.Parallel.Strategies hiding (parPair)
import Control.Monad.Parallel as Mp
import System.IO.Error
import System.Environment(getArgs, getProgName)

type MatBox = (Int, Int, String) -- x, y, val
type Pattern = (Int, Int) -- row, col, diag pattern

instance {-# OVERLAPPING #-} Show MatBox => Show (Int, Int, String)
  where
    show (x, y, val) = (show val)

parPair :: Strategy (a, b)
parPair (a, b) = do
  a' <- rpar a
  b' <- rpar b
  return (a', b')

-- Gen List for dimensions -- 
genList :: Int -> [Int]
genList x = [1..x]

-- In case players without predefined bingo boards, the players get assigned random bingo cards -- 
getShuffles :: RandomGen gen => [Int] -> Int -> gen -> [[Int]]
getShuffles mat player gen = ((Split.chunksOf 5 cards) !! ( player `mod` 5 ) )
                          where
                            cards = map (\cur -> shuffle' cur (length cur) gen) spurs
                            spurs = take 50 (map (\cr -> shuffle' cr (length cr) ngn) permt)
                            permt = take 1000 (List.permutations mat)
                            ngn = mkStdGen 40

getMatrix :: [Int] -> Mat.Matrix Int
getMatrix a = Mat.fromList x x a where x = round ( sqrt (fromIntegral $ length a) )

-- Get Bingo Board for a player -- 
getBoard :: Int -> Mat.Matrix Int -> Mat.Matrix MatBox
getBoard dim matc = Mat.matrix dim dim $ \( i, j ) -> ( i, j, show (Mat.getElem i j matc)  )

-- For striking, we anyways have to traverse the whole matrix, since we go row by row -- 
searchRow :: (Int, Int) -> Mat.Matrix MatBox -> Int -> (Bool, Int)
searchRow (x, y) card val | y > (Mat.ncols card) = (False, y)
                          | vv == (show val) = (True, y)
                          | otherwise = searchRow (x, y + 1) card val
                           where (rr, cc, vv) = (Mat.getElem x y card)

searchMatrix :: Mat.Matrix MatBox -> (Int, Int) -> Int -> (Mat.Matrix MatBox, (Int, Int))
searchMatrix card (x, y) val | state = ( (Mat.setElem (x, col, (show 'X')) (x, col) card), (x, col) )
                             | otherwise = searchMatrix card (x + 1, y) val
                               where (state, col) = searchRow (x, y) card val

-- Row, Col, Diag checks -- 
checkRow :: (Int, Int) -> Mat.Matrix MatBox -> Bool
checkRow (x, y) card | y > (Mat.ncols card) = True
                     | vv /= (show 'X') = False
                     | otherwise = checkRow (x, y + 1) card
                      where (rr, cc, vv) = (Mat.getElem x y card)

checkCol :: (Int, Int) -> Mat.Matrix MatBox -> Bool
checkCol (x, y) card | x > (Mat.nrows card) = True
                     | vv /= (show 'X') = False
                     | otherwise = checkCol (x + 1, y) card
                      where (rr, cc, vv) = (Mat.getElem x y card)

checkDiag :: (Int, Int) -> Mat.Matrix MatBox -> Bool
checkDiag (x, y) card | x > (Mat.nrows card) = True
                      | vv /= (show 'X') = False
                      | otherwise = checkDiag ( x+1, y+1) card
                        where (rr, cc, vv) = (Mat.getElem x y card)

-- Check Sol or checkWin -- 
checkSolution :: (Int, Int) -> Mat.Matrix MatBox -> Bool
checkSolution (x, y) card | x == y = cr || cc || (checkDiag (1, 1) card) 
                          | otherwise = cr || cc
                          where (cr, cc) = ((checkRow (x, 1) card), (checkCol (1, y) card)) `using` parPair -- parallelized

strikeBox :: Int -> Mat.Matrix MatBox -> (Mat.Matrix MatBox, (Int, Int))
strikeBox val card = searchMatrix card (1, 1) val

printMatrix :: (Int, Mat.Matrix MatBox) -> String
printMatrix (player, mat) = "Player " ++ (show player) ++ "\n" ++ (Mat.prettyMatrix mat)

-- playGame for each number the game host calls -- 
playGame :: [Int] -> [Mat.Matrix MatBox] -> IO ()
playGame [] _ = putStrLn "Game Over"
playGame (x:xs) pcards = do
  let rest = map (\card -> strikeBox x card) pcards `using` parList rseq -- parallelized
  let ncards = map (\(mat, (x, y)) -> mat ) rest
  let sols = map (\(mat, (x, y)) -> checkSolution (x, y) mat ) rest `using` parList rseq -- parallelized
  let winners = filter (\(player, state) -> state) $ zip [1..( (length pcards) ::Int )] sols
  -- putStrLn $ "Game Host calls " ++ (show x) ++ "!!"
  -- Prelude.mapM_ (putStrLn . printMatrix) (zip [1..( (length pcards) ::Int )] ncards) -- uncomment for output
  if null winners then
    playGame xs ncards
  else
    if (length winners) == 1 then
      die $ "BINGO!!! *** Player " ++ (show ( fst (head winners) ) ) ++ " Won!! ***"
    else 
      die $  "Its a Draw!!"

--- just added for test cases ---
readLines :: FilePath -> IO [String]
readLines = fmap lines . readFile

makeInteger :: [String] -> [Int]
makeInteger = map read
----        *******       -----

main :: IO ()
main = do args <- getArgs
          case args of 
            [d, p] -> do
              let dimensions = read d :: Int
              let players = read p :: Int
              rng <- newStdGen
              let mcards = map (\player -> (getShuffles (genList (dimensions * dimensions) ) player (mkStdGen player)) !! ( player `mod` 5 ) ) [1..(players :: Int)]
              let ghcard = shuffle' (genList (dimensions * dimensions)) (dimensions * dimensions) rng
              putStrLn "These are the current players assigned bingo cards"
              Prelude.mapM_ (putStrLn . Mat.prettyMatrix . getMatrix) mcards
              let boards = map (\x -> getBoard dimensions (getMatrix x)) mcards
              playGame ghcard boards
            [pl1f, pl2f, ghf, d] -> do
              p1b <- readLines pl1f
              p2b <- readLines pl2f
              ghb <- readLines ghf
              let pl1m = makeInteger p1b
              let pl2m = makeInteger p2b
              let ghm = makeInteger ghb
              let players = 2
              let dimensions = read d :: Int
              let mcards = [pl1m, pl2m]
              let ghcard = ghm
              Prelude.mapM_ (putStrLn . Mat.prettyMatrix . getMatrix) mcards
              let boards = map (\x -> getBoard dimensions (getMatrix x)) mcards
              playGame ghcard boards
            [players, ghf, d] -> do
              scards <- readLines players
              let hycards = map words scards
              let dimensions = read d :: Int
              let mcards = map makeInteger hycards
              let players = (length mcards) ::Int
              ghb <- readLines ghf
              let ghm = makeInteger ghb
              let ghcard = ghm
              putStrLn "These are the current players assigned bingo cards"
              -- Prelude.mapM_ (putStrLn . Mat.prettyMatrix . getMatrix) mcards -- uncomment for initial card outputs
              let boards = map (\x -> getBoard dimensions (getMatrix x)) mcards
              putStrLn "Start game"
              playGame ghcard boards
            _ -> do pn <- getProgName
                    die $ "Usage: " ++ pn ++ " <board_dimensions> <total_players> or <player1_file> <player2_file> <gamehost_file> <dimensions> or <players_file> <gamehost_file> <dimensions>"
   `catchIOError` \ e -> die $ case ioeGetFileName e of
          Just fn | isDoesNotExistError e -> fn ++ ": No such file"
                  | isPermissionError e -> fn ++ ": Permission denied"
          _       | isUserError e -> "Usage: ./bingo <board_dimensions> <total_players>"
                  | otherwise -> show e
    



